// TreeIterator utility
// Copyright (c) 2001 Alexander Ananiev & MyArch.com.  All rights reserved.

package com.myarch.treeiterator;

import java.util.Iterator;

/**
 *  This class deals with enumeration representing a level of a tree,
 * i.e., all children of the parent.
 * <p> Provides the abstraction for <code>TreeIterator</code> and encapsulates
 * all the details related to the navigation on the level.
 * <p> Class is made final for performance.
 *  @author Alexander Ananiev
 */
final class LevelIterator implements Iterator,
  ObjectPool.Factory, TreeConstants {

  // singleton -- fake enumeration for the root
  static private RootIter rootIter = new RootIter();

  private ObjectPool iterPool;
  private TreeNodeAdapter nodeAdapter;
  private int  index =-1;
  private int depth = 0;
  private LevelIterator parent;
  private Object node;
  private Iterator iter;
  private int maxDepth;

  /**
   * Creates iterator for the root of the tree.
   */
  LevelIterator( Object rootNode, TreeNodeAdapter nodeAdapter ) {
    this.nodeAdapter = nodeAdapter;
    // root is not counted by iterator
    setRoot( rootNode );
  }

  /**
   * Creates empty iterator.
   * <code>setLevel</code> or <code>setRoot</code>must be called after that.
   * Empty constructor is needed for object pooling.
   */
  private LevelIterator() {

  }

  /**
   * Resets the iterator to the root state.
   * Fake enumeration object is used to simulate empty enumeration for the root
   * level.
   */
  void setRoot( Object node ) {
    this.node = node;
    iter = rootIter;
    parent = null;
    index = 0;
    depth = 0;
    maxDepth = DEPTH_UNLIMITED;
    iterPool = new ObjectPool( this );
    iterPool.add( this );
  }

  /**
   * Sets level data. Used only by <code>nextLevel</code> in conjunction with
   * parameterless constructor.
   * <p> <code>TreeIerator.next()</code> must be called right after that, otherwise
   * iterator doesn't have a current node.
   */
  private void setLevel( ObjectPool iterPool, Iterator iter, LevelIterator parent,
    int depth, int maxDepth, TreeNodeAdapter nodeAdapter) {

    this.iterPool = iterPool;
    this.nodeAdapter = nodeAdapter;
    this.iter = iter;
    this.parent = parent;
    index = -1;
    this.depth = depth;
    this.maxDepth = maxDepth;
    // no current node
    node = null;

  }


  /**
   * <code>ObjectFactory</code> method. Used by <code>TObjectPool</code>
   * to create new instance of the <code>TLevelIterator</code>.
   */
  public Object create() {
    return new LevelIterator();
  }

  /**
   * Returns current node
   */
  Object current() {
    return node;
  }

  /**
   * Returns index of the node within enumeration. Index starts with 0.
   */
  int getIndex(){
    return index;
  }

  /**
   *  Sets the initial depth of the traversal.
   */
  void setDepth( int depth ) {
    this.depth = depth;
  }

  /**
   * Returns the depth of the level being traversed. Root has depth 0.
   */
  int getDepth() {
    return depth;
  }
  /**
   * Sets max traversal depth allowed for this level.
   * <code>TLevelIterator</code> simply keeps it so it can be changed dynamicaly
   * based on level.
   */
  void setMaxDepth( int maxDepth ) {
    this.maxDepth = maxDepth;
  }

  /**
   * Returns <code>true</code> if the max allowed level is reached.
   */
  boolean isMaxDepth() {
    return (maxDepth != DEPTH_UNLIMITED && depth >= maxDepth );
  }

  /**
   * Returns max allowed depth set for this level
   */
  public int getMaxDepth() {
     return maxDepth;
  }

  /**
   * Path-through to iterator's calls.
   */
  public Object next() {
    node = iter.next();
    ++index;
    return node;
  }

  /**
   * Path-through to iterator's calls.
   */
  public boolean hasNext() {
    return iter.hasNext();
  }

   /**
    * Remove is not supported
    */
   public void remove() throws UnsupportedOperationException {
     throw new UnsupportedOperationException("Remove is not supported by LevelIterator");
   }



  /**
   * Returns <code>true</code> if current node has children
   */
  public boolean hasChildren() {
    return nodeAdapter.hasChildren( node );
  }

  /**
   * Makes children of the current node new level.
   * Returns new iterator for this level.
   * Sets itself as a parent for the new level
   */
  LevelIterator nextLevel() {

    LevelIterator aParent = this;
    // It can be parent only if some elements left.
    // This way we can return to the prev. level without looping and resume iteration.
    if ( ! hasNext() ) {
      iterPool.release();
      aParent = this.parent;
    }

    // call object pool and get new iterator
    LevelIterator newIter = (LevelIterator)iterPool.allocate();
    newIter.setLevel(iterPool, nodeAdapter.getChildren( node ), aParent,
      depth+1, maxDepth, nodeAdapter);

    return newIter;
  }

  /**
   * Returns <code>true</code> if this level has parent.
   * Tree traversal stops if there is no parent.
   */
  boolean hasParent() {
    return (parent!=null);
  }

  /**
   * Move to the parent level. Current iterator should
   * be returned to the pool.
   */
  LevelIterator toParent() {
    iterPool.release();
    return parent;
  }

  /**
   * Dummy <code>Iterator</code> to be used for the root
   * of the tree. Root is never processed and it doesn't have siblings.
   */
  private static class RootIter implements Iterator{

    /**
     * Returns <code>null</code> at the attempt to advance.
     */
    public Object next() {
      return null;
    }
    /**
     * Always returns <code>false</code> -- always empty.
     */
    public boolean hasNext() {
      return false;
    }

   /**
    * Remove is not supported
    */
   public void remove() throws UnsupportedOperationException {
     throw new UnsupportedOperationException("Remove is not supported by RootIter");
   }


  }

}