// TreeIterator utility
// Copyright (c) 2001 Alexander Ananiev & MyArch.com.  All rights reserved.


package com.myarch.treeiterator;

import java.util.Iterator;

/**
 * Iterates over the tree. Preorder traversal is used, i.e.,
 * node is processed before its children.
 * <br>Depth of traversal can be limited by <code>setMaxDepth</code>.
 * <br>Branches can be traversed selectively using <code>skipChildren()</code>.
 * <br>Unlike <code>java.util.Iterator</code>, the tree iterator points directly at the element (tree node),
 * not BETWEEN elements. It means that, for example, if <code>next()</code> is
 * called before processing, root won't be processed.
 * <code>current()</code> can be used to obtain the current element at
 * any time (except right after <code>moveToChildren()</code> was called).
 * <p>Usage:
 * <br>Instantiation:
 * <code>
 * <br>TreeIterator tIterator = new TreeIterator ( rootNode , new DOMAdapter() );
 * </code>
 * <br>Tree traversal INCLUDING the root:
 * <pre><code>
 * for( Object node = tIterator.current(); node!=null; node=tIterator.next() ){
 *     // processing...
 * }
 * </code></pre>
 * <br>Tree traversal EXCLUDING the root:
 * <pre><code>
 * while( (node=(Node)tIterator.next()) != null )  {
 *     // processing...
 * }
 * </code></pre>
 * <br>Tree traversal using <code>hasNext()</code>:
 * <pre><code>
 * while( tIterator.hasNext() ) {
 *     Node node=(Node)tIterator.next();
 *     // processing...
 *  }
 * </code></pre>
 * <br>Note that <code>hasNext()</code> implementation is not very efficient for a
 * tree structure from the performance standpoint, so refrain from using it for large trees.
 *
 * @author Alexander Ananiev
 */
public class TreeIterator implements TreeConstants, Iterator {

  private LevelIterator levelIter;
  private boolean skipChildren =false;

  // constructors

  /**
   * Creates iterator and sets the root of a tree and tree adapter.
   *
   * @param root          root of the tree; note that <code>next()</code> moves to the first child, not to
   * the root
   * @param nodeAdapter   tree node adapter
   */
  public TreeIterator( Object root, TreeNodeAdapter nodeAdapter ) {
    levelIter = new LevelIterator( root, nodeAdapter );
  }

  /**
   * Creates iterator and sets the root of a tree, tree adapter and depth of the
   * traversal.
   *
   * @param root          root of the tree
   * @param nodeAdapter   tree node adapter
   * @param maxDepth      depth of the traversal; 0 -- traverse only the root; 1-- traverse root and
   * childrenl; 2-- traverse root, children and grandchildren and so forth
   *
   */
   public TreeIterator( Object root, TreeNodeAdapter nodeAdapter, int maxDepth) {
    levelIter = new LevelIterator( root, nodeAdapter );
    setMaxDepth( maxDepth );
   }


   /**
    * This method is defined to comply with <code>Iterator</code> but it is not supported.
    * @throws UnsupportedOperationException whenever it is called
    */
   public void remove() throws UnsupportedOperationException {
     throw new UnsupportedOperationException("Remove is not supported by TreeIterator");
   }


  /**
   * Sets the root of the tree.
   * @param root          root of the tree; note that <code>next()</code> moves to the first child, not to
   * the root
   */
  public void setRoot( Object node ) {
    levelIter.setRoot( node );
  }

  /**
   * Returns the current node. After iterator is created, root becomes current.
   * Every <code>next()</code> call moves current to the next node.
   * @return current node
   */
  public Object current() {
    return levelIter.current();
  }

  /**
   * Sets the maximum depth of the traversal.
   * @param maxDepth      depth of the traversal; 0 -- traverse only the root; 1-- traverse root and
   * childrenl; 2-- traverse root, children and grandchildren and so forth
   */
  public void setMaxDepth( int maxDepth ){
    levelIter.setMaxDepth( maxDepth );
  }

 /**
   * Returns the maximum depth of the traversal.
   *
   * @return maximum allowed traversal depth set by <code>setMaxDepth()</code>.
   */
  public int getMaxDepth( ){
    return levelIter.getMaxDepth( );
  }

  /**
   * Moves iterator to the children of the node
   * regardless of current level depth setting.
   * <p>Iterator doesn't have the current node after this call.
   * It is positioned before the first child.
   * <code>next()</code> must be called to position it to the first child.
   * The entire branch will be processed. Previous
   * max depth is restored after the branch is processed.
   */

  public void moveToChildren() {
    moveToChildren( DEPTH_UNLIMITED );
  }

  /**
   * Moves iterator to the children of the node
   * regardless of current level depth setting.
   * <p> Iterator doesn't have current node after this call.
   * It is positioned before the first child.
   * <code>Next()</code> must be called to position to the first child.
   * <p>It is the alternative and more sophisticated approach to using <code>skipChildren()</code>.
   * <p> <code>setLevelDepth</code> does the same thing.
   * @param depth New depth that applies to all children of the current node. Previous
   * depth is restored after the branch is processed. This depth setting is differeent
   * from the maximum depth of
   * the traversal as it only applies to the current branch. It overrides the maximum depth previously set
   * by <code>setMaxDepth()</code>.
   */
  public void moveToChildren( int depth ) {
    if ( levelIter.hasChildren() ) {
      levelIter = levelIter.nextLevel();
      levelIter.setMaxDepth( depth );
      skipChildren = true;
    }
  }

  /**
   * Same with <code>moveToChildren</code>.
   */
  public void setLevelDepth( int depth ) {
    moveToChildren( depth );
  }


  /**
   * After this method is called, <code>next()</code> will skip children of the current node
   * and move to the next sibling instead.
   */
  public void skipChildren() {
    skipChildren = true;
  }

  /**
   * Returns <code>true</code> if the maximum allowed level is reached.
   * @return true if this is the level set by <code>setMaxDepth()</code>
   */
  public boolean isMaxDepth() {
    return levelIter.isMaxDepth();
  }

  /**
   * Returns the depth of the level being traversed.
   * @return  depth of the current node; root has depth 0, children -- 1, grandchildren -- 2 and so on
   */
  public int getDepth() {
    return levelIter.getDepth();
  }

  /**
   * Sets the initial depth value. It does not move the iterator, it simply
   * forces it to start counting depth not from 0, but from this value. I.e.,
   * if initial depth is set to 2, root depth would be 2, depth of its children would be 3 and so on.
   * @param depth initial depth
   */
  public void setDepth( int depth ) {
    levelIter.setDepth( depth );
  }

  /**
   * Returns index of the of the current node among siblings.
   * Index starts with 0.
   * @return index of the current node
   */
  public int getIndex() {
    return levelIter.getIndex();
  }


  /**
   * Returns <code>true</code> if not all elements of the tree have
   * been traversed.
   * <p> It is not recommended to use this method, since this check is
   * performed anyway inside <code>next()</code>. It is more efficient to use
   * <code>next()</code> and check for <code>null</code>.
   * @return <code>true</code> if end of tree is not reached
   */
  public boolean hasNext() {
    return( hasChildren() ||  levelIter.hasNext() || levelIter.hasParent() );
  }

  /**
   * Returns <code>true</code> if current node has children and iterator
   * is allowed to traverse them.
   * @return <code>true</code> if current node has children
   */
  public boolean hasChildren() {
    return (!levelIter.isMaxDepth() && ! skipChildren && levelIter.hasChildren() );
  }

  /**
   * Returns the next element of a tree.
   * Preorder traversal is perfromed.
   * <br>Note that unlike <code>java.util.Iterator</code> it does not
   * raise exception when the end of the tree is reached, <code>null</code> is
   * returned instead.
   * @return the next node or <code>null</code> when there is no more nodes to traverse
   */
  public Object next() {
    // go to the next level only if maxDepth would not be exceeded
    if ( hasChildren() ) {
        levelIter=levelIter.nextLevel();
    }

    skipChildren = false;

    Object nextElt=null;
    // if there are more siblings at the current level
    if ( levelIter.hasNext() )
      nextElt=levelIter.next();
    // the entire level has been processed--go to the parent level
    else if (levelIter.hasParent() ) {
      levelIter=levelIter.toParent();
      nextElt=levelIter.next();
    }

    return nextElt;
  }

}
